/***************************************************************************
 * e2ls - ls program for ext2 file systems
 *
 * Copyright (C) 1995 Claus Tondering, ct@login.dknet.dk
 * Copyright (C) 1995 Peeter Joot
 *
 * This file may be redistributed under the terms of the GNU Public License.
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <time.h>

#include <linux/types.h>

#include "linux/ext2_fs.h"
#include "linux/stat.h"

#include "ext2fs/ext2fs.h"
#include "ext2fs/ext2_err.h"
#include "getopt.h"
#include "e2err.h"
#include "dev.h"

extern io_manager nt_io_manager;

int aflag, dflag, iflag, lflag, tflag, rflag;

struct fileinfo {
	ino_t	inode;
	unsigned char	name[80];
	struct ext2_inode e2ino;
} *list;
int listix;
int maxlist;

void outino(struct fileinfo *);


/**********************************************************************
 * myproc is a callback routine which is called once for each entry
 * in the directory
 **********************************************************************/

static int myproc(struct ext2_dir_entry *dirent,
				  int	offset,
				  int	blocksize,
				  char	*buf,
				  void	*private)
{
	if (!list) {
		list = malloc(100*sizeof(struct fileinfo));
		if (!list)
			return EXT2_ET_BADMEM;
		maxlist = 100;
	}

	if (dirent->name[0]=='.' && !aflag)
		return 0;

	list[listix].inode=dirent->inode;
	strncpy(list[listix].name,dirent->name,dirent->name_len);
	list[listix].name[dirent->name_len] = 0;

	listix++;
	if (listix==maxlist) {
		list = realloc(list, (maxlist+100) * sizeof(struct fileinfo));
		if (!list) {
			fprintf(stderr,"Cannot allocate memory\n");
			return DIRENT_ABORT;
		}
		maxlist += 100;
	}
	return 0;
}


/**********************************************************************
 * compare is used by qsort() to compare to list entries
 **********************************************************************/

int
compare(const void *ee1, const void *ee2)
{
	int res;
	const struct fileinfo *e1 = ee1, *e2 = ee2;

	if (tflag)
		res = e1->e2ino.i_mtime < e2->e2ino.i_mtime ? 1 : -1;
	else
		res = strcmp(e1->name, e2->name);

	return rflag ? -res : res;
}


/**********************************************************************
 * usage prints usage information and exits
 **********************************************************************/

void
usage()
{
	fprintf(stderr,
"SYNOPSIS\n"
"\tE2LS [-adilhtr] [pathname]\n"
"\n"
"DESCRIPTION\n"
"\tThe pathname must identify an ext2 directory or file. The E2LS\n"
"\tprogram provides a list of the contents of that directory in a\n"
"\tmanner similar to the Linux ls program.\n"
"\n"
"\tThe following options are supported:\n"
"\n"
"-a\tInclude file names starting with . in listing.\n"
"\n"
"-d\tList only pathname, even if it is a directory.\n"
"\n"
"-i\tInclude inode number in listing.\n"
"\n"
"-l\tProduce a long listing.\n"
"\n"
"-t\tSort by modification time.\n"
"\n"
"-r\tReverse the sort order.\n"
"\n"
"-h\tPrint this message.\n"
);
	exit(1);
}


/**********************************************************************
 * main routine
 **********************************************************************/

char * prog_name;

main(int argc, char **argv)
{
	int err, i, c;
	ext2_filsys fs;
	ino_t ino;
	struct ext2_inode e2ino;
	char *filename;
	e2cwd_t cwd;

	prog_name = argv[0];
	initialize_ext2_error_table();

	opterr = 0;
	while ((c=getopt(argc, argv, "adihltr")) != -1) {
		switch (c) {
			case 'a': aflag++; break;
			case 'd': dflag++; break;
			case 'i': iflag++; break;
			case 'l': lflag++; break;
			case 't': tflag++; break;
			case 'r': rflag++; break;
			case 'h':
			case '?': usage();
		}
	}

	if (argc==optind)
		filename = ".";
	else if (argc!=optind+1)
		usage();
	else
		filename = argv[optind];

        err = cwdenv_lookup(&cwd);
        if (err)
		e2_err(err,0);

	/* Open file system */
	err = ext2fs_open((const char *)&cwd, 0, 0, 0, nt_io_manager, &fs);
	if (err)
		e2_err(EXT2_ET_CANT_OPEN_FS,err);

	/* Lookup specified name */
	err = ext2fs_namei(fs, 2, cwd.cwdino, filename, &ino);
	if (err)
		e2_err(EXT2_ET_NO_FILE,err);

	/* Read specified inode */
	err = ext2fs_read_inode(fs, ino, &e2ino);
	if (err)
		e2_err(EXT2_ET_CANT_READ_IN_INFO,err);

	/* Is it a directory? */
	if (!S_ISDIR(e2ino.i_mode) || dflag) {
		struct fileinfo fi;
		fi.inode = ino;
		strcpy(fi.name, filename);
		fi.e2ino = e2ino;
		outino(&fi);
	}
	else {
		/* Loop through all directory entries */
		err = ext2fs_dir_iterate(fs, ino, 0, 0, myproc, 0);
		if (err)
			e2_err(EXT2_ET_CANT_READ_DIR,err);

		if (lflag || tflag)
			for (i=0; i<listix; i++) {
				err = ext2fs_read_inode(fs, list[i].inode, &list[i].e2ino);
				if (err)
					e2_err(EXT2_ET_CANT_READ_IN_INFO,err);
			}


		qsort((void*)list, (size_t)listix, (size_t)sizeof(list[0]),
			compare);

		for (i=0; i<listix; i++)
			outino(&list[i]);
	}

	return 0;
}		


/**********************************************************************
 * outino writes one line of directory information
 **********************************************************************/

void
outino(struct fileinfo *fi)
{
	char *mtimestring;
	time_t now;

	time(&now);

	if (iflag)
		printf("%5d ",fi->inode);

	if (lflag) {
		if (S_ISDIR(fi->e2ino.i_mode))
			printf("d");
		else if (S_ISREG(fi->e2ino.i_mode))
			printf("-");
		else if (S_ISLNK(fi->e2ino.i_mode))
			printf("l");
		else if (S_ISCHR(fi->e2ino.i_mode))
			printf("c");
		else if (S_ISBLK(fi->e2ino.i_mode))
			printf("b");
		else if (S_ISFIFO(fi->e2ino.i_mode))
			printf("p");
		else if (S_ISSOCK(fi->e2ino.i_mode))
			printf("s");
		else
			printf("?");

		printf("%c%c%c%c%c%c%c%c%c ",
			   		fi->e2ino.i_mode&S_IRUSR ? 'r' : '-',
			   		fi->e2ino.i_mode&S_IWUSR ? 'w' : '-',
			   		fi->e2ino.i_mode&S_IXUSR ? 'x' : '-',
			   		fi->e2ino.i_mode&S_IRGRP ? 'r' : '-',
			   		fi->e2ino.i_mode&S_IWGRP ? 'w' : '-',
			   		fi->e2ino.i_mode&S_IXGRP ? 'x' : '-',
			   		fi->e2ino.i_mode&S_IROTH ? 'r' : '-',
			   		fi->e2ino.i_mode&S_IWOTH ? 'w' : '-',
			   		fi->e2ino.i_mode&S_IXOTH ? 'x' : '-');

		printf("%2d  %4d %4d ",
			   fi->e2ino.i_links_count, fi->e2ino.i_uid, fi->e2ino.i_gid);

		if (S_ISCHR(fi->e2ino.i_mode) || S_ISBLK(fi->e2ino.i_mode))
			printf("%3d, %3d ", fi->e2ino.i_block[0]>>8, fi->e2ino.i_block[0]&0xff);
		else
			printf("%8d ", fi->e2ino.i_size);

		mtimestring = ctime(&fi->e2ino.i_mtime);

		if (fi->e2ino.i_mtime < (unsigned long)(now-6*30*24*60*60)) /* File more than 6 months old */
			printf("%.6s  %.4s ",mtimestring+4, mtimestring+20);
		else
			printf("%.12s ",mtimestring+4);
	}

	printf("%s",fi->name);
	
	if (lflag) {
		if (S_ISLNK(fi->e2ino.i_mode))
			printf(" -> %s",(char*)fi->e2ino.i_block);
	}
	printf("\n");
}

